package explorviz.hpc_monitoring.filter.reconstruction;

import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.lmax.disruptor.EventHandler;
import com.lmax.disruptor.RingBuffer;
import com.lmax.disruptor.dsl.Disruptor;

import explorviz.hpc_monitoring.disruptor.RecordArrayEvent;
import explorviz.hpc_monitoring.disruptor.RecordEvent;
import explorviz.hpc_monitoring.filter.counting.CountingThroughputFilter;
import explorviz.hpc_monitoring.filter.reduction.TracePatternSummarizationFilter;
import explorviz.hpc_monitoring.reader.IPeriodicTimeSignalReceiver;
import explorviz.hpc_monitoring.reader.TimedReader;
import explorviz.hpc_monitoring.record.HostApplicationMetadata;
import explorviz.hpc_monitoring.record.IRecord;
import explorviz.hpc_monitoring.record.Trace;
import explorviz.hpc_monitoring.record.event.AbstractOperationEvent;

public final class TraceReconstructionFilter implements
		EventHandler<RecordArrayEvent>, IPeriodicTimeSignalReceiver {
	private static final int OUTPUT_MESSAGE_BUFFER_SIZE = 256;

	private static final CountingThroughputFilter counter = new CountingThroughputFilter(
			"Reconstructed traces per second");

	private final long maxTraceTimeout;

	private final Map<Long, TraceBuffer> traceId2trace = new TreeMap<Long, TraceBuffer>();

	private final RingBuffer<RecordArrayEvent> ringBuffer;
	private IRecord[] outputBuffer = new IRecord[OUTPUT_MESSAGE_BUFFER_SIZE];
	private int outputBufferIndex = 0;

	public TraceReconstructionFilter(final long maxTraceTimeout,
			final EventHandler<RecordEvent> endReceiver) {
		this.maxTraceTimeout = maxTraceTimeout;

		final ExecutorService exec = Executors.newCachedThreadPool();
		final Disruptor<RecordArrayEvent> disruptor = new Disruptor<RecordArrayEvent>(
				RecordArrayEvent.EVENT_FACTORY, 32, exec);

		@SuppressWarnings("unchecked")
		final EventHandler<RecordArrayEvent>[] eventHandlers = new EventHandler[1];
		eventHandlers[0] = new TracePatternSummarizationFilter(1 * 1000,
				endReceiver);
		disruptor.handleEventsWith(eventHandlers);
		ringBuffer = disruptor.start();

		new TimedReader(1 * 1000, this).start();
	}

	@Override
	public void periodicTimeSignal(final long timestamp) {
		checkForTimeouts(timestamp); // TODO comes from other thread -
										// synchronize!
		flushOutputBuffer();
	}

	private void checkForTimeouts(final long timestamp) {
		final long traceTimeout = timestamp - maxTraceTimeout;
		for (final Entry<Long, TraceBuffer> entry : traceId2trace.entrySet()) {
			final TraceBuffer traceBuffer = entry.getValue();
			if (traceBuffer.getMaxLoggingTimestamp() <= traceTimeout) {
				sendOutInvalidTrace(traceBuffer.toTrace());
				// TODO remove from traceId2trace
			}
		}
	}

	private void sendOutValidTrace(final Trace trace) {
		putInRingBuffer(trace);
	}

	private void sendOutInvalidTrace(final Trace trace) {
		// putInRingBuffer(trace); // TODO
		System.out.println("Invalid trace: "
				+ trace.getTraceEvents()[0].getTraceId());
	}

	private void putInRingBuffer(final IRecord message) {
		counter.inputObjects(message);
		synchronized (this) { // TODO remove
			outputBuffer[outputBufferIndex++] = message;

			if (outputBufferIndex == OUTPUT_MESSAGE_BUFFER_SIZE) {
				flushOutputBuffer();
			}
		}
	}

	private void flushOutputBuffer() {
		synchronized (this) { // TODO remove
			if (outputBufferIndex > 0) {
				final long hiseq = ringBuffer.next();
				final RecordArrayEvent valueEvent = ringBuffer.get(hiseq);
				valueEvent.setValues(outputBuffer);
				ringBuffer.publish(hiseq);

				outputBuffer = new IRecord[OUTPUT_MESSAGE_BUFFER_SIZE]; // TODO
				// object
				// reusage?
				outputBufferIndex = 0;
			}
		}
	}

	@Override
	public void onEvent(final RecordArrayEvent event, final long sequence,
			final boolean endOfBatch) throws Exception {
		for (final IRecord record : event.getValues()) { // TODO save length in
															// event
			if (record != null) {
				final AbstractOperationEvent abstractOperationEvent = ((AbstractOperationEvent) record);

				final long traceId = abstractOperationEvent.getTraceId();
				final TraceBuffer traceBuffer = getBufferForTraceId(
						abstractOperationEvent.getTraceId(),
						event.getMetadata());
				traceBuffer.insertEvent(abstractOperationEvent);

				if (traceBuffer.isFinished()) {
					traceId2trace.remove(traceId);
					sendOutValidTrace(traceBuffer.toTrace());
				}
			}
		}
	}

	private TraceBuffer getBufferForTraceId(final long traceId,
			final HostApplicationMetadata metadata) {
		TraceBuffer traceBuffer = traceId2trace.get(traceId);
		if (traceBuffer == null) {
			traceBuffer = new TraceBuffer(); // TODO dont create new - keep old
												// ones and reset!
			traceBuffer.setTrace(metadata); // TODO reuse...
			traceId2trace.put(traceId, traceBuffer);
		}
		return traceBuffer;
	}

	public void terminate() {
		for (final Object entry : traceId2trace.values()) {
			if (entry instanceof TraceBuffer) {
				sendOutInvalidTrace(((TraceBuffer) entry).toTrace());
			}
		}
		traceId2trace.clear();
	}
}