package explorviz.hpc_monitoring.reader;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.lmax.disruptor.EventHandler;
import com.lmax.disruptor.RingBuffer;
import com.lmax.disruptor.dsl.Disruptor;

import explorviz.hpc_monitoring.disruptor.RecordArrayEvent;
import explorviz.hpc_monitoring.disruptor.RecordEvent;
import explorviz.hpc_monitoring.filter.counting.CountingThroughputFilter;
import explorviz.hpc_monitoring.filter.reconstruction.TraceReconstructionFilter;
import explorviz.hpc_monitoring.record.HostApplicationMetadata;
import explorviz.hpc_monitoring.record.IRecord;
import explorviz.hpc_monitoring.record.event.normal.AfterFailedOperationEvent;
import explorviz.hpc_monitoring.record.event.normal.AfterOperationEvent;
import explorviz.hpc_monitoring.record.event.normal.BeforeOperationEvent;

public final class TCPReader implements IPeriodicTimeSignalReceiver {
	private static final int MESSAGE_BUFFER_SIZE = 131072;
	private static final int OUTPUT_MESSAGE_BUFFER_SIZE = 16384;

	// Settings
	private final int listeningPort;
	private boolean active = true;

	private ServerSocketChannel serversocket;
	private HostApplicationMetadata hostApplicationMetadata;

	// Buffers
	private final RingBuffer<RecordArrayEvent> ringBuffer;
	private IRecord[] outputBuffer = new IRecord[OUTPUT_MESSAGE_BUFFER_SIZE];
	private int outputBufferIndex = 0;

	private static final CountingThroughputFilter counter = new CountingThroughputFilter(
			"Records per second");

	private final static Map<Integer, String> stringRegistry = new TreeMap<Integer, String>();
	private final static List<byte[]> waitingForStringMessages = new ArrayList<byte[]>(
			1024);

	public TCPReader(final int listeningPort,
			final EventHandler<RecordEvent> endReceiver) {
		this.listeningPort = listeningPort;

		final ExecutorService exec = Executors.newCachedThreadPool();
		final Disruptor<RecordArrayEvent> disruptor = new Disruptor<RecordArrayEvent>(
				RecordArrayEvent.EVENT_FACTORY, 16, exec);

		@SuppressWarnings("unchecked")
		final EventHandler<RecordArrayEvent>[] eventHandlers = new EventHandler[1];
		eventHandlers[0] = new TraceReconstructionFilter(1 * 1000 * 1000,
				endReceiver);
		disruptor.handleEventsWith(eventHandlers);
		ringBuffer = disruptor.start();

		new TimedReader(1 * 1000, this).start();
	}

	@Override
	public void periodicTimeSignal(final long timestamp) {
		synchronized (this) { // TODO remove
			flushOutputBuffer();
		}
	}

	public final void read() {
		try {
			open();
			while (active) {
				// TODO only one connection!
				final ByteBuffer buffer = ByteBuffer
						.allocateDirect(MESSAGE_BUFFER_SIZE);
				final SocketChannel socketChannel = serversocket.accept();
				while ((socketChannel.read(buffer)) != -1) {
					buffer.flip();
					messagesfromByteArray(buffer);
				}

				serversocket.close();
			}
		} catch (final IOException ex) {
			System.out.println("Error in read() " + ex.toString());
		} finally {
			try {
				serversocket.close();
			} catch (final IOException e) {
				System.out.println("Error in read()" + e.toString());
			}
		}
	}

	private final void open() throws IOException {
		serversocket = ServerSocketChannel.open();
		serversocket.socket().bind(new InetSocketAddress(listeningPort));
		System.out.println("listening on port " + listeningPort);
	}

	public final void terminate(final boolean error) {
		System.out.println("Shutdown of TCPReader requested.");
		active = false;
	}

	private final void messagesfromByteArray(final ByteBuffer buffer) {
		while (buffer.remaining() > 0) {
			buffer.mark();
			try {
				final byte clazzId = buffer.get();

				switch (clazzId) {
					case HostApplicationMetadata.CLAZZ_ID: {
						readInTraceMetadata(buffer);
						break;
					}
					case BeforeOperationEvent.CLAZZ_ID: {
						readInBeforeOperationEvent(buffer);
						break;
					}
					case AfterFailedOperationEvent.CLAZZ_ID: {
						readInAfterFailedOperationEvent(buffer);
						break;
					}
					case AfterOperationEvent.CLAZZ_ID: {
						readInAfterOperationEvent(buffer);
						break;
					}
					case 4: {
						final int mapId = buffer.getInt();
						final int stringLength = buffer.getInt();

						final byte[] stringByteArray = new byte[stringLength];

						buffer.get(stringByteArray);

						addToRegistry(mapId, new String(stringByteArray));
						break;
					}
					default: {
						System.out.println("unknown class id " + clazzId
								+ " at offset " + (buffer.position() - 4));
					}
				}
			} catch (final BufferUnderflowException e) {
				buffer.reset();
				buffer.compact();
				return;
			}
		}

		buffer.clear();
	}

	private final void readInTraceMetadata(final ByteBuffer buffer) {
		final int hostnameId = buffer.getInt();
		final int applicationId = buffer.getInt();

		final String hostname = getStringFromRegistry(hostnameId);
		final String application = getStringFromRegistry(applicationId);

		if ((hostname != null) && (application != null)) {
			hostApplicationMetadata = new HostApplicationMetadata(hostname,
					application);
		} else {
			final byte[] message = new byte[HostApplicationMetadata.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position()
					- HostApplicationMetadata.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}
	}

	private final void readInBeforeOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);

		if (operation != null) {
			putInRingBuffer(new BeforeOperationEvent(timestamp, traceId,
					orderIndex, operation));
		} else {
			final byte[] message = new byte[BeforeOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position()
					- BeforeOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}
	}

	private final void readInAfterFailedOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();
		final int causeId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);
		final String cause = getStringFromRegistry(causeId);

		if ((operation != null) && (cause != null)) {
			putInRingBuffer(new AfterFailedOperationEvent(timestamp, traceId,
					orderIndex, operation, cause));
		} else {
			final byte[] message = new byte[AfterFailedOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position()
					- AfterFailedOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}
	}

	private final void readInAfterOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);
		if (operation != null) {
			putInRingBuffer(new AfterOperationEvent(timestamp, traceId,
					orderIndex, operation));
		} else {
			final byte[] message = new byte[AfterOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position()
					- AfterOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}

	}

	private final void putInRingBuffer(final IRecord message) {
		counter.inputObjects(message);

		synchronized (this) { // TODO remove
			outputBuffer[outputBufferIndex++] = message;
			if (outputBufferIndex == OUTPUT_MESSAGE_BUFFER_SIZE) {
				flushOutputBuffer();
			}
		}
	}

	private void flushOutputBuffer() {
		if (outputBufferIndex > 0) {
			final long hiseq = ringBuffer.next();
			final RecordArrayEvent valueEvent = ringBuffer.get(hiseq);
			valueEvent.setValues(outputBuffer);
			valueEvent.setMetadata(hostApplicationMetadata);
			ringBuffer.publish(hiseq);

			outputBuffer = new IRecord[OUTPUT_MESSAGE_BUFFER_SIZE]; // TODO
			// object
			// reusage?
			outputBufferIndex = 0;
		}
	}

	private final void putInWaitingMessages(final byte[] message) {
		waitingForStringMessages.add(message); // TODO kill messages if too long
												// in queue
	}

	private final void checkWaitingMessages() {
		final List<byte[]> localWaitingList = new ArrayList<byte[]>();
		for (final byte[] waitingMessage : waitingForStringMessages) {
			localWaitingList.add(waitingMessage);
		}
		waitingForStringMessages.clear();

		for (final byte[] waitingMessage : localWaitingList) {
			final ByteBuffer buffer = ByteBuffer.wrap(waitingMessage);
			final byte waitingMessageClazzId = buffer.get();
			switch (waitingMessageClazzId) {
				case HostApplicationMetadata.CLAZZ_ID:
					readInTraceMetadata(buffer);
					break;
				case BeforeOperationEvent.CLAZZ_ID:
					readInBeforeOperationEvent(buffer);
					break;
				case AfterFailedOperationEvent.CLAZZ_ID:
					readInAfterFailedOperationEvent(buffer);
					break;
				case AfterOperationEvent.CLAZZ_ID:
					readInAfterOperationEvent(buffer);
					break;
				default:
					break;
			}
		}
	}

	private final void addToRegistry(final int key, final String value) {
		stringRegistry.put(key, value);

		// System.out.println("put key " + key + " value " + value);

		checkWaitingMessages();
	}

	private final String getStringFromRegistry(final int id) {
		return stringRegistry.get(id);
	}
}
